#!/usr/bin/env python3

# source ./activate_test_env.sh
# subprojects/libinsane-gobject/examples/scan.py

import os
import PIL.Image
import sys
import traceback

import gi
gi.require_version('Libinsane', '1.0')
from gi.repository import GObject  # noqa: E402


from gi.repository import Libinsane  # noqa: E402


#! [ExampleLogger]
class ExampleLogger(GObject.GObject, Libinsane.Logger):
    def do_log(self, lvl, msg):
        if lvl <= Libinsane.LogLevel.INFO:
            return
        print("{}: {}".format(lvl.value_nick, msg))
        sys.stdout.flush()
#! [ExampleLogger]


#! [ExampleListGetDevice]
def get_device(api, dev_id=None):
    if dev_id is None:
        print("Looking for scan devices ...")
        devs = api.list_devices(Libinsane.DeviceLocations.ANY)
        print("Found {} devices".format(len(devs)))
        for dev in devs:
            print("[{}] : [{}]".format(dev.get_dev_id(), dev.to_string()))
        dev_id = devs[0].get_dev_id()

    print("Will use device {}".format(dev_id))
    dev = api.get_device(dev_id)
    print("Using device {}".format(dev.get_name()))
    return dev
#! [ExampleListGetDevice]


#! [ExampleListGetScanSource]
def get_source(dev, source_name):
    print("Looking for scan sources ...")
    sources = dev.get_children()
    print("Available scan sources:")
    for src in sources:
        print("- {}".format(src.get_name()))
        if src.get_name() == source_name:
            source = src
            break
    else:
        if source_name is None:
            source = sources[0] if len(sources) > 0 else dev
        elif source_name == "root":
            source = dev
        else:
            print("Source '{}' not found".format(source_name))
            sys.exit(2)
    print("Will use scan source {}".format(source.get_name()))
    return source
#! [ExampleListGetScanSource]


#! [ExampleSetOption]
def set_opt(item, opt_name, opt_value):
    try:
        print("Setting {} to {}".format(opt_name, opt_value))
        opts = item.get_options()
        opts = {opt.get_name(): opt for opt in opts}
        if opt_name not in opts:
            print("Option '{}' not found".format(opt_name))
            return
        print("- Old {}: {}".format(opt_name, opts[opt_name].get_value()))
        print("- Allowed values: {}".format(opts[opt_name].get_constraint()))
        set_flags = opts[opt_name].set_value(opt_value)
        print("- Set flags: {}".format(set_flags))
        opts = item.get_options()
        opts = {opt.get_name(): opt for opt in opts}
        print("- New {}: {}".format(opt_name, opts[opt_name].get_value()))
    except Exception as exc:
        print("Failed to set {} to {}: {}".format(
            opt_name, opt_value, str(exc)
        ))
        traceback.print_exc()
    finally:
        print("")
#! [ExampleSetOption]


#! [ExampleListOptions]
def list_opts(item):
    opts = item.get_options()
    print("Options:")
    for opt in opts:
        try:
            print("- {}={} ({})".format(
                opt.get_name(), opt.get_value(), opt.get_constraint()
            ))
        except Exception as exc:
            print("Failed to read option {}: {}".format(
                opt.get_name(), str(exc)
            ))
    print("")
#! [ExampleListOptions]


#! [ExampleRawToImg]
def raw_to_img(params, img_bytes):
    fmt = params.get_format()
    assert(fmt == Libinsane.ImgFormat.RAW_RGB_24)
    (w, h) = (
        params.get_width(),
        int(len(img_bytes) / 3 / params.get_width())
    )
    print("Mode: RGB : Size: {}x{}".format(w, h))
    return PIL.Image.frombuffer("RGB", (w, h), img_bytes, "raw", "RGB", 0, 1)
#! [ExampleRawToImg]


#! [ExampleScan]
def scan(source, output_file):
    session = source.scan_start()

    try:
        page_nb = 0
        while not session.end_of_feed() and page_nb < 20:
            # Do not assume that all the pages will have the same size !
            scan_params = session.get_scan_parameters()
            print("Expected scan parameters: {} ; {}x{} = {} bytes".format(
                  scan_params.get_format(),
                  scan_params.get_width(), scan_params.get_height(),
                  scan_params.get_image_size()))
            total = scan_params.get_image_size()

            img = []
            r = 0
            if output_file is not None:
                out = output_file.format(page_nb)
            else:
                out = None
            print("Scanning page {} --> {}".format(page_nb, out))
            while not session.end_of_page():
                data = session.read_bytes(128 * 1024)
                data = data.get_data()
                img.append(data)
                r += len(data)
                print("Got {} bytes => {}/{} bytes".format(
                    len(data), r, total)
                )
            img = b"".join(img)
            print("Got {} bytes".format(len(img)))
            if out is not None:
                print("Saving page as {} ...".format(out))
                if scan_params.get_format() == Libinsane.ImgFormat.RAW_RGB_24:
                    img = raw_to_img(scan_params, img)
                    img.save(out, format="PNG")
                else:
                    print("Warning: output format is {}".format(
                        scan_params.get_format()
                    ))
                    with open(out, 'wb') as fd:
                        fd.write(img)
            page_nb += 1
            print("Page {} scanned".format(page_nb))
        if page_nb == 0:
            print("No page in feeder ?")
    finally:
        session.cancel()
#! [ExampleScan]


def main():
    if False:
        os.environ['LIBINSANE_NORMALIZER_ALL_OPTS_ON_ALL_SOURCES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_BMP2RAW'] = "0"
        os.environ['LIBINSANE_NORMALIZER_CLEAN_DEV_DESCS'] = "0"
        os.environ['LIBINSANE_NORMALIZER_MIN_ONE_SOURCE'] = "0"
        os.environ['LIBINSANE_NORMALIZER_OPT_ALIASES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_RAW24'] = "0"
        os.environ['LIBINSANE_NORMALIZER_RESOLUTION'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SAFE_DEFAULTS'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_NAMES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_NODES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_TYPES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_CACHE'] = "0"
        os.environ['LIBINSANE_WORKAROUND_CHECK_CAPABILITIES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_DEDICATED_THREAD'] = "0"
        os.environ['LIBINSANE_WORKAROUND_LAMP'] = "0"
        os.environ['LIBINSANE_WORKAROUND_ONE_PAGE_FLATBED'] = "0"
        os.environ['LIBINSANE_WORKAROUND_OPT_NAMES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_OPT_VALUES'] = "0"

#! [ExampleSetLogger]
    Libinsane.register_logger(ExampleLogger())
#! [ExampleSetLogger]

    if len(sys.argv) <= 1 or (sys.argv[1] == "-h" or sys.argv[1] == "--help"):
        print(
            "Syntax: {}"
            " <PNG output file>"
            " [<scan dev id> [<scan source name>]]".format(sys.argv[0])
        )
        sys.exit(1)

    output_file = sys.argv[1]
    dev_id = None
    source = None
    source_name = None
    if len(sys.argv) > 2:
        dev_id = sys.argv[2]
    if len(sys.argv) > 3:
        source_name = sys.argv[3]

    print("Will write the scan result into {}".format(output_file))

#! [ExampleInit]
    api = Libinsane.Api.new_safebet()
#! [ExampleInit]

    dev = get_device(api, dev_id)
    try:
        source = get_source(dev, source_name)

        list_opts(source)

        # set the options
#! [ExampleOptsToSet]
        set_opt(source, 'resolution', 300)
        # set_opt(source, 'mode', "LineArt")
        # set_opt(source, 'depth', 1)
#! [ExampleOptsToSet]

        print("Scanning ...")
        scan(source, output_file)
        print("Scan done")
    finally:
        dev.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# source ./activate_test_env.sh
# subprojects/libinsane-gobject/examples/list_devices.py

import gi
gi.require_version('Libinsane', '1.0')
from gi.repository import GObject  # noqa: E402

#! [Logger]
from gi.repository import Libinsane  # noqa: E402


class ExampleLogger(GObject.GObject, Libinsane.Logger):
    def do_log(self, lvl, msg):
        if lvl <= Libinsane.LogLevel.WARNING:
            return
        print("{}: {}".format(lvl.value_nick, msg))


def main():
    Libinsane.register_logger(ExampleLogger())
    #! [Logger]
    api = Libinsane.Api.new_safebet()
    print("Looking for devices ...")
    devs = api.list_devices(Libinsane.DeviceLocations.ANY)
    print("Found {} devices".format(len(devs)))
    for d in devs:
        try:
            dev = api.get_device(d.get_dev_id())
            print("|")
            print("|-- {} ({} ; {})".format(
                d.get_dev_id(), d.to_string(), dev.get_name()
            ))
            try:
                for child in dev.get_children():
                    print("|   |-- {}".format(child.get_name()))
            finally:
                dev.close()
        except Exception as exc:
            print("ERROR: failed to open device {}: {}".format(
                d.get_dev_id(), str(exc)
            ))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# source ./activate_test_env.sh
# subprojects/libinsane-gobject/examples/list_options.py

import os

import gi
gi.require_version('Libinsane', '1.0')
from gi.repository import GObject  # noqa: E402

from gi.repository import Libinsane  # noqa: E402


class ExampleLogger(GObject.GObject, Libinsane.Logger):
    def do_log(self, lvl, msg):
        if lvl <= Libinsane.LogLevel.DEBUG:
            return
        print("{}: {}".format(lvl.value_nick, msg))


def main():
    if True:
        os.environ['LIBINSANE_NORMALIZER_ALL_OPTS_ON_ALL_SOURCES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_BMP2RAW'] = "0"
        os.environ['LIBINSANE_NORMALIZER_CLEAN_DEV_DESCS'] = "0"
        os.environ['LIBINSANE_NORMALIZER_MIN_ONE_SOURCE'] = "0"
        os.environ['LIBINSANE_NORMALIZER_OPT_ALIASES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_RAW24'] = "0"
        os.environ['LIBINSANE_NORMALIZER_RESOLUTION'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SAFE_DEFAULTS'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_NAMES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_NODES'] = "0"
        os.environ['LIBINSANE_NORMALIZER_SOURCE_TYPES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_CACHE'] = "0"
        os.environ['LIBINSANE_WORKAROUND_CHECK_CAPABILITIES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_DEDICATED_PROCESS'] = "0"
        os.environ['LIBINSANE_WORKAROUND_DEDICATED_THREAD'] = "0"
        os.environ['LIBINSANE_WORKAROUND_LAMP'] = "0"
        os.environ['LIBINSANE_WORKAROUND_ONE_PAGE_FLATBED'] = "0"
        os.environ['LIBINSANE_WORKAROUND_OPT_NAMES'] = "0"
        os.environ['LIBINSANE_WORKAROUND_OPT_VALUES'] = "0"

    Libinsane.register_logger(ExampleLogger())
    api = Libinsane.Api.new_safebet()

    devs = api.list_devices(Libinsane.DeviceLocations.ANY)
    print("Found {} devices".format(len(devs)))
    for dev in devs:
        print("")
        print("")
        print(dev.to_string())

        try:
            dev = api.get_device(dev.get_dev_id())
            children = dev.get_children()
            print("|-- Found {} sources".format(len(children)))

            for item in [dev] + children:
                print("|")
                print("|-- {} ({})".format(
                    item.get_name(),
                    "device" if item == dev else "source"
                ))
                opts = item.get_options()
                for opt in opts:
                    print("|   |-- {}".format(opt.get_name()))
                    print("|   |   |-- Title: {}".format(opt.get_title()))
                    print("|   |   |-- Description: {}".format(opt.get_desc()))
                    print("|   |   |-- Capabilities: {}".format(
                        opt.get_capabilities())
                    )
                    print("|   |   |-- Unit: {}".format(opt.get_value_unit()))
                    print("|   |   |-- Constraint type: {}".format(
                          opt.get_constraint_type()))
                    print("|   |   |-- Constraint: {}".format(
                        opt.get_constraint())
                    )
                    if opt.is_readable():
                        try:
                            print("|   |   |-- Value: {}".format(
                                opt.get_value())
                            )
                        except Exception as exc:
                            print("|   |   |-- Value: [{}]".format(str(exc)))
                    else:
                        print("|   |   |-- Value: (unavailable)")
            print("")
            dev.close()
        except Exception as exc:
            print("EXCEPTION: {}".format(exc))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# source ./activate_test_env.sh
# subprojects/libinsane-gobject/examples/test_hw.py <directory>

import os
import re
import sys

import PIL.Image

import gi
gi.require_version('Libinsane', '1.0')
from gi.repository import GObject  # noqa: E402
from gi.repository import Libinsane  # noqa: E402


class Logger(GObject.GObject, Libinsane.Logger):
    def do_log(self, lvl, msg):
        if lvl < Libinsane.LogLevel.WARNING:
            return
        print("{}: {}".format(lvl.value_nick, msg))


def get_devices(api):
    print("Looking for scan devices ...")
    dev_descs = api.list_devices(Libinsane.DeviceLocations.ANY)
    print("Found {} devices".format(len(dev_descs)))
    for dev in dev_descs:
        print("[{}] : [{}]".format(dev.get_dev_id(), dev.to_string()))
    devs = []
    for dev in dev_descs:
        print("")
        dev_id = dev.get_dev_id()
        devs.append(dev_id)
    return devs


def get_source(dev, source_name):
    print("Looking for scan sources ...")
    sources = dev.get_children()
    print("Found {} scan sources:".format(len(sources)))

    for src in sources:
        print("- {}".format(src.get_name()))

    for src in sources:
        if src.get_name().lower() == source_name.lower():
            print("Will use scan source {}".format(src.get_name()))
            return src

    for src in sources:
        if src.get_name().lower().startswith(source_name.lower()):
            print("Will use scan source {}".format(src.get_name()))
            return src

    raise Exception("Source '{}' not found".format(source_name))


def list_opts(item):
    opts = item.get_options()
    print("Options:")
    for opt in opts:
        try:
            print("- {}={} ({})".format(
                opt.get_name(), opt.get_value(), opt.get_constraint()
            ))
        except Exception as exc:
            print("Failed to read option {}: {}".format(
                opt.get_name(), str(exc)
            ))


def set_opt(item, opt_name, opt_values):
    print("Setting {} to {}".format(opt_name, opt_values))
    opts = item.get_options()
    opts = {opt.get_name(): opt for opt in opts}
    if opt_name not in opts:
        raise Exception("Option '{}' not found".format(opt_name))
    print("- Old {}: {}".format(opt_name, opts[opt_name].get_value()))
    print("- Allowed values: {}".format(opts[opt_name].get_constraint()))
    for opt_value in opt_values:
        if opt_value not in opts[opt_name].get_constraint():
            continue
        opts[opt_name].set_value(opt_value)
        opts = item.get_options()
        opts = {opt.get_name(): opt for opt in opts}
        print("- New {}: {}".format(opt_name, opts[opt_name].get_value()))
        return
    raise Exception(
        "Failed to set option {}: No value match constraint".format(opt_name)
    )


def raw_to_img(params, img_bytes):
    fmt = params.get_format()
    assert(fmt == Libinsane.ImgFormat.RAW_RGB_24)
    (w, h) = (
        params.get_width(),
        int(len(img_bytes) / 3 / params.get_width())
    )
    mode = "RGB"
    print("Mode: {} : Size: {}x{}".format(mode, w, h))
    return PIL.Image.frombuffer(mode, (w, h), img_bytes, "raw", mode, 0, 1)


def scan(source, output_file):
    session = source.scan_start()

    scan_params = session.get_scan_parameters()
    print("Expected scan parameters: {} ; {}x{} = {} bytes".format(
          scan_params.get_format(),
          scan_params.get_width(), scan_params.get_height(),
          scan_params.get_image_size()))

    try:
        page_nb = 0

        assert(not session.end_of_feed())

        img = []
        print("Scanning page --> {} ...".format(output_file))
        while not session.end_of_page():
            data = session.read_bytes(32 * 1024)
            data = data.get_data()
            img.append(data)
        img = b"".join(img)
        img = raw_to_img(scan_params, img)
        print("Saving page as {} ...".format(output_file))
        img.save(output_file)
        print("Page scanned".format(page_nb))

        assert(session.end_of_page())
        # TODO(Jflesch): HP ScanJet 4300C
        # assert(session.end_of_feed())
    finally:
        session.cancel()


def clean_filename(filename):
    return re.sub("[^0-9a-zA-Z.]", "_", filename)


def main():
    Libinsane.register_logger(Logger())

    if len(sys.argv) <= 1 or (sys.argv[1] == "-h" or sys.argv[1] == "--help"):
        print("Syntax: {} <output directory>".format(sys.argv[0]))
        sys.exit(1)

    output_dir = sys.argv[1]

    print("Will write the scan result into {}/".format(output_dir))
    os.mkdir(output_dir)

    api = Libinsane.Api.new_safebet()

    print("Looking for devices ...")

    devs = get_devices(api)
    if len(devs) <= 0:
        print("No device found")
        sys.exit(1)

    for dev_id in devs:
        print("Will use device {}".format(dev_id))
        for t in range(0, 3):
            print("- Test {}".format(t))
            dev = api.get_device(dev_id)
            try:
                print("Using device {}".format(dev.get_name()))
                output_file = clean_filename(
                    dev.get_name() + "_" + str(t) + ".jpeg"
                )
                output_file = os.path.join(output_dir, output_file)

                print("Looking for source flatbed ...")
                src = get_source(dev, "flatbed")
                list_opts(src)

                # set the options
                set_opt(src, 'resolution', [150, 200, 300])
                if t == 2:
                    set_opt(src, 'mode', ['LineArt'])
                else:
                    set_opt(src, 'mode', ['Color'])

                print("Scanning ...")
                scan(src, output_file)
                print("Scan done")
            finally:
                dev.close()

    print("All scan done")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import json
import os
import subprocess
import sys


def get_cflags(compile_commands):
    for cfile in compile_commands:
        cflags = cfile['command']
        cflags = cflags.split(" ")
        out = []
        for f in cflags:
            if f == "":
                continue
            if f[0] == "'" and f[-1] == "'":
                f = f[1:-1]
            if f[:2] == "-D" or f[:2] == "-I":
                out.append(f)
        yield (out, cfile['file'])


if __name__ == "__main__":
    with open(sys.argv[1], 'r') as fd:
        compile_commands = json.load(fd)
    os.chdir(os.path.dirname(sys.argv[1]))
    for (cflags, filepath) in get_cflags(compile_commands):
        if "libinsane-gobject" in filepath:
            # gobject macros raise many warnings --> do not check
            continue
        cflags += [
            '-O2', '-Wsparse-all', '-Wsparse-error', '-Wno-non-pointer-null'
        ]
        # WORKAROUND(Jflesch):
        # TODO(Jflesch): Why do I need to do that ?
        cflags += ['-D_Float32=float']
        cflags += ['-D_Float32x=float']
        cflags += ['-D_Float64=double']
        cflags += ['-D_Float64x=double']
        cflags += ['-D_Float128=double']
        cflags += ['-D_Float128x=double']
        cflags += ['-D__ARM_PCS_VFP']
        print("Running sparse on {}".format(filepath))
        cmd = ['sparse', filepath] + cflags
        r = subprocess.run(cmd)
        if r.returncode != 0:
            print("Command was: {}".format(cmd))
            sys.exit(r.returncode)
